/* Copyright (C) 2012 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <stdlib.h>

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#if LUA_VERSION_NUM == 501
#define lua_rawlen lua_objlen
#endif

struct array {
    enum type {
	DOUBLE,
	FLOAT,
	ULONG,
	LONG,
	UINT,
	INT,
	USHORT,
	SHORT,
	UCHAR,
	CHAR
    } type;

    enum {
	FREE_NOTHING,
	FREE_SIZE,
	FREE_BOTH
    } free;
    
    int rank, *size;
    union values {
	void *any;
	double *doubles;
	float *floats;
	unsigned long *ulongs;
	signed long *longs;
	unsigned int *uints;
	signed int *ints;
	unsigned short *ushorts;
	signed short *shorts;
	unsigned char *uchars;
	signed char *chars;
    } values;
};

static int __index (lua_State *L);
static int __newindex (lua_State *L);
static int __gc (lua_State *L);
static int __len (lua_State *L);
static int __ipairs (lua_State *L);
static int __tostring (lua_State *L);

static size_t sizeof_element (enum type type)
{
    switch (type) {
    case DOUBLE:
	return sizeof (double);
    case FLOAT:
	return sizeof (float);
    case ULONG:
	return sizeof (unsigned long);
    case LONG:
	return sizeof (long);
    case UINT:
	return sizeof (unsigned int);
    case INT:
	return sizeof (int);
    case USHORT:
	return sizeof (unsigned short);
    case SHORT:
	return sizeof (short);
    case UCHAR:
	return sizeof (unsigned char);
    case CHAR:
	return sizeof (char);
    }

    return 0;
}

static void write_element (struct array *array, int i, lua_Number value)
{
    switch (array->type) {
    case DOUBLE:
	array->values.doubles[i] = value;
	break;
    case FLOAT:
	array->values.floats[i] = value;
	break;
    case ULONG:
	array->values.ulongs[i] = value;
	break;
    case LONG:
	array->values.longs[i] = value;
	break;
    case UINT:
	array->values.uints[i] = value;
	break;
    case INT:
	array->values.ints[i] = value;
	break;
    case USHORT:
	array->values.ushorts[i] = value;
	break;
    case SHORT:
	array->values.shorts[i] = value;
	break;
    case UCHAR:
	array->values.uchars[i] = value;
	break;
    case CHAR:
	array->values.chars[i] = value;
	break;
    }
}

static lua_Number read_element (struct array *array, int i)
{
    switch (array->type) {
    case DOUBLE:
	return (lua_Number)array->values.doubles[i];
    case FLOAT:
	return (lua_Number)array->values.floats[i];
    case ULONG:
	return (lua_Number)array->values.ulongs[i];
    case LONG:
	return (lua_Number)array->values.longs[i];
    case UINT:
	return (lua_Number)array->values.uints[i];
    case INT:
	return (lua_Number)array->values.ints[i];
    case USHORT:
	return (lua_Number)array->values.ushorts[i];
    case SHORT:
	return (lua_Number)array->values.shorts[i];
    case UCHAR:
	return (lua_Number)array->values.uchars[i];
    case CHAR:
	return (lua_Number)array->values.chars[i];
    }

    return 0;
}

static void *reference_element (struct array *array, int i)
{
    switch (array->type) {
    case DOUBLE:
	return &array->values.doubles[i];
    case FLOAT:
	return &array->values.floats[i];
    case ULONG:
	return &array->values.ulongs[i];
    case LONG:
	return &array->values.longs[i];
    case UINT:
	return &array->values.uints[i];
    case INT:
	return &array->values.ints[i];
    case USHORT:
	return &array->values.ushorts[i];
    case SHORT:
	return &array->values.shorts[i];
    case UCHAR:
	return &array->values.uchars[i];
    case CHAR:
	return &array->values.chars[i];
    }

    return NULL;
}

static void copy_elements (struct array *from, struct array *to, int i, int n)
{
    switch (from->type) {
    case DOUBLE:
	memcpy(&to->values.doubles[i], from->values.doubles, n * sizeof(to->values.doubles[0]));
	break;
    case FLOAT:
	memcpy(&to->values.floats[i], from->values.floats, n * sizeof(to->values.floats[0]));
	break;
    case ULONG:
	memcpy(&to->values.ulongs[i], from->values.ulongs, n * sizeof(to->values.ulongs[0]));
	break;
    case LONG:
	memcpy(&to->values.longs[i], from->values.longs, n * sizeof(to->values.longs[0]));
	break;
    case UINT:
	memcpy(&to->values.uints[i], from->values.uints, n * sizeof(to->values.uints[0]));
	break;
    case INT:
	memcpy(&to->values.ints[i], from->values.ints, n * sizeof(to->values.ints[0]));
	break;
    case USHORT:
	memcpy(&to->values.ushorts[i], from->values.ushorts, n * sizeof(to->values.ushorts[0]));
	break;
    case SHORT:
	memcpy(&to->values.shorts[i], from->values.shorts, n * sizeof(to->values.shorts[0]));
	break;
    case UCHAR:
	memcpy(&to->values.uchars[i], from->values.uchars, n * sizeof(to->values.uchars[0]));
	break;
    case CHAR:
	memcpy(&to->values.chars[i], from->values.chars, n * sizeof(to->values.chars[0]));
	break;
    }
}

static void dump (lua_State *L, struct array *array, int l, int b)
{
    int i, d;

    if (lua_rawlen (L, -1) != array->size[l]) {
	lua_pushfstring (L,
			 "Inconsistent array structure (subarray "
			 "at depth %d should have %d elements but "
			 "has %d).",
			 l, array->size[l], lua_rawlen (L, -1));
	lua_error (L);
    }	
	
    for (i = l + 1, d = 1;
	 i < array->rank;
	 d *= array->size[i], i += 1);

    for (i = 0 ; i < array->size[l] ; i += 1) {
	lua_rawgeti (L, -1, i + 1);
	
	if (lua_type (L, -1) == LUA_TNUMBER) {
	    write_element(array, b + i, lua_tonumber (L, -1));
	} else {
	    dump (L, array, l + 1, b + i * d);
	}
	
	lua_pop (L, 1);
    }
}

static void construct (lua_State *L, struct array *array, int reference)
{
    *(struct array *)lua_newuserdata(L, sizeof(struct array)) = *array;

    lua_newtable (L);
    
    if (reference > 0) {
	lua_pushvalue(L, reference);
	lua_rawseti(L, -2, 1);
    }
    
    lua_pushstring(L, "__ipairs");
    lua_pushcfunction(L, (lua_CFunction)__ipairs);
    lua_settable(L, -3);
    lua_pushstring(L, "__pairs");
    lua_pushcfunction(L, (lua_CFunction)__ipairs);
    lua_settable(L, -3);
    lua_pushstring(L, "__len");
    lua_pushcfunction(L, (lua_CFunction)__len);
    lua_settable(L, -3);
    lua_pushstring(L, "__index");
    lua_pushcfunction(L, (lua_CFunction)__index);
    lua_settable(L, -3);
    lua_pushstring(L, "__newindex");
    lua_pushcfunction(L, (lua_CFunction)__newindex);
    lua_settable(L, -3);
    lua_pushstring(L, "__tostring");
    lua_pushcfunction(L, (lua_CFunction)__tostring);
    lua_settable(L, -3);
    lua_pushstring(L, "__gc");
    lua_pushcfunction(L, (lua_CFunction)__gc);
    lua_settable(L, -3);
    lua_setmetatable(L, -2);
}

static int __gc (lua_State *L)
{
    struct array *array;
    
    array = lua_touserdata(L, 1);

    if (array->free > FREE_NOTHING) {
	free(array->size);
    }

    if (array->free > FREE_SIZE) {
	free(array->values.any);
    }

    return 0;
}

static int __tostring (lua_State *L)
{
    struct array *array;
    int i;
    
    array = lua_touserdata(L, 1);

    lua_pushfstring (L, "array[%d", array->size[0]);

    for (i = 1 ; i < array->rank ; i += 1) {
	lua_pushfstring (L, ", %d", array->size[i]);
    }

    lua_pushstring (L, "]");

    lua_concat (L, i + 1);
    
    return 1;
}

static int __len (lua_State *L)
{
    struct array *array;
    
    array = lua_touserdata(L, 1);

    lua_pushinteger (L, array->size[0]);
    
    return 1;
}

static int __next (lua_State *L)
{
    struct array *array;
    int k;
    
    array = lua_touserdata(L, 1);
    k = lua_tointeger (L, 2);

    lua_pop (L, 1);
    
    if (k < array->size[0]) {
	lua_pushinteger (L, k + 1);
	lua_pushinteger (L, k + 1);
	lua_gettable (L, 1);

	return 2;
    } else {
	lua_pushnil (L);

	return 1;
    }
}

static int __ipairs (lua_State *L)
{
    lua_pushcfunction (L, __next);
    lua_insert (L, 1);
    lua_pushnil (L);

    return 3;
}

static int __index (lua_State *L)
{
    struct array *array;
    int i;
    
    array = lua_touserdata(L, 1);
    i = lua_tointeger (L, 2);

    if (i > 0 && i <= array->size[0]) {
	if (array->rank == 1) {
	    lua_pushnumber (L, read_element(array, lua_tointeger (L, 2) - 1));
	} else {
	    struct array subarray;
	    int j, d;

	    for (j = 1, d = 1;
		 j < array->rank;
		 d *= array->size[j], j += 1);

	    subarray.type = array->type;
	    subarray.free = FREE_NOTHING;
	    subarray.rank = array->rank - 1;
	    subarray.size = &array->size[1];
	    subarray.values.any = reference_element(array, (i - 1) * d);

	    construct (L, &subarray, 1);
	}
    } else {
	lua_pushnil (L);
    }

    return 1;
}

static int __newindex (lua_State *L)
{
    struct array *array;
    int i;
    
    array = lua_touserdata(L, 1);
    i = lua_tointeger (L, 2);

    if (i > 0 && i <= array->size[0]) {
	if (array->rank == 1) {
	    write_element (array, lua_tointeger (L, 2) - 1, lua_tonumber (L, 3));
	} else if (lua_istable (L, 3)) {
	    struct array subarray;
	    int j, r, d;
	    
	    for (r = 0;
		 lua_type (L, -1) == LUA_TTABLE;
		 r += 1, lua_rawgeti (L, -1, 1)) {
		if (array->size[r + 1] != lua_rawlen (L, -1)) {
		    lua_pushstring (L, "Array sizes don't match.");
		    lua_error (L);
		}
	    }

	    if (r != array->rank - 1) {
		lua_pushstring (L, "Array dimensions don't match.");
		lua_error (L);		
	    }

	    lua_settop (L, 3);

	    for (j = 1, d = 1;
		 j < array->rank;
		 d *= array->size[j], j += 1);

	    subarray.type = array->type;
	    subarray.free = FREE_NOTHING;
	    subarray.rank = array->rank - 1;
	    subarray.size = &array->size[1];
	    subarray.values.any = reference_element(array, (i - 1) * d);

	    dump (L, &subarray, 0, 0);
	} else {
	    struct array *subarray;
	    int j, d;

	    subarray = lua_touserdata(L, 3);

	    for (j = 1, d = 1;
		 j < array->rank;
		 d *= array->size[j], j += 1);

	    if (subarray->rank != array->rank - 1) {
		lua_pushstring (L, "Array dimensions don't match.");
		lua_error (L);
	    }

	    if (subarray->type != array->type) {
		lua_pushstring (L, "Array types don't match.");
		lua_error (L);
	    }

	    for (j = 1 ; j < array->rank ; j += 1) {
		if (subarray->size[j - 1] != array->size[j]) {
		    lua_pushstring (L, "Array sizes don't match.");
		    lua_error (L);
		}
	    }
	    
	    copy_elements(subarray, array, (i - 1) * d, d);
	}
    } else {
	lua_pushstring (L, "Index out of array bounds.");
	lua_error (L);
    }

    return 0;
}

static int fromtable (lua_State *L, struct array *array)
{
    int i, l, r, h;

    luaL_checktype (L, -1, LUA_TTABLE);

    h = lua_gettop (L);
    
    for (r = 0;
	 lua_type (L, -1) == LUA_TTABLE;
	 r += 1, lua_rawgeti (L, -1, 1));

    lua_settop (L, h);

    if (array->rank == 0) {
	array->rank = r;
	array->size = calloc (r, sizeof(int));
    } else if (r != array->rank) {
	lua_pushstring (L, "Initialization from incompatible table (dimensions don't match).");
	lua_error (L);
    }

    for (i = 0, l = 1 ; i < r ; i += 1) {
	int l_0;
	
	l_0 = lua_rawlen (L, -1);
	l *= l_0;

	if (array->size[i] == 0) {
	    array->size[i] = l_0;
	} else if (l_0 != array->size[i]) {
	    lua_pushstring (L, "Initialization from incompatible table (sizes don't match).");
	    lua_error (L);
	}
	
	lua_rawgeti (L, -1, 1);
    }

    lua_settop (L, h);

    array->free = FREE_BOTH;
    array->values.any = malloc (l * sizeof_element (array->type));

    dump (L, array, 0, 0);
    construct (L, array, 0);

    return 1;
}

static void fromstring (lua_State *L, struct array *array)
{
    int i, l;

    for (i = 0, l = 1 ; i < array->rank ; l *= array->size[i], i += 1);

    array->free = FREE_SIZE;
    array->values.any = (void *)lua_tostring (L, -1);

    l *= sizeof_element (array->type);
	
    if (l != lua_rawlen (L, -1)) {
	lua_pushfstring (L, "Invalid array data length (should be %d bytes but is %d bytes).", l, lua_rawlen (L, -1));
	lua_error (L);
    }

    construct (L, array, lua_gettop(L));
}

static int create (lua_State *L)
{
    struct array array;
    int i, j;

    for (i = 1 ; lua_type (L, i) == LUA_TNUMBER ; i += 1);

    if (lua_type (L, i) != LUA_TSTRING &&
	lua_type (L, i) != LUA_TTABLE) {
	lua_pushfstring (L,
			 "Initialization from incompatible value "
			 "(expected string or table, got a %s).",
			 lua_typename (L, lua_type (L, i)));
	lua_error (L);
    }

    array.type = lua_tointeger (L, lua_upvalueindex(1));
    
    if (i < 2) {
	if (lua_type(L, i) == LUA_TSTRING) {
	    array.rank = 1;
	    
	    array.size = malloc (sizeof (int));
	    array.size[0] = lua_rawlen(L, i) / sizeof_element (array.type);
	} else {
	    array.rank = 0;
	    array.size = NULL;
	}	    
    } else {
	array.rank = i - 1;
	array.size = malloc ((i - 1) * sizeof (int));
	
	for (j = 0 ; j < i - 1 ; j += 1) {
	    array.size[j] = lua_tonumber (L, j + 1);
	}
    }

    if (lua_type(L, i) == LUA_TSTRING) {
	fromstring (L, &array);
    } else {
	fromtable (L, &array);
    }
    
    return 1;
}

static int cast (lua_State *L)
{
    struct array cast, *array;
    int i, j, l, m;

    for (i = 1 ; lua_type (L, i) == LUA_TNUMBER ; i += 1);

    if (i < 2) {
	lua_pushstring (L, "Invalid array dimensions.");
	lua_error (L);
    }

    luaL_checktype (L, i, LUA_TUSERDATA);

    array = lua_touserdata (L, i);
    cast.size = malloc ((i - 1) * sizeof (int));;
    
    for (j = 0, l = 1 ; j < i - 1 ; j += 1) {
	cast.size[j] = lua_tonumber (L, j + 1);
	l *= cast.size[j];
    }

    for (j = 0, m = 1 ; j < array->rank ; j += 1) {
	m *= array->size[j];
    }

    if (l != m) {
	lua_pushstring (L, "Incompatible cast dimensions.");
	lua_error (L);
    }
	
    cast.type = array->type;
    cast.rank = i - 1;
    cast.values.any = array->values.any;
    cast.free = FREE_SIZE;

    construct (L, &cast, i);
    
    return 1;
}

static int copy (lua_State *L)
{
    struct array *array, copy;
    int i, d;
    
    array = lua_touserdata(L, 1);

    copy.type =	array->type;
    copy.rank = array->rank;
    copy.free = FREE_BOTH;

    copy.size = malloc (array->rank * sizeof (int));
    memcpy(copy.size, array->size,array->rank * sizeof (int));
    
    for (i = 0, d = 1;
	 i < array->rank;
	 d *= array->size[i], i += 1);

    copy.values.any = malloc (d * sizeof_element (array->type));
    memcpy(copy.values.any, array->values.any,
	   d * sizeof_element (array->type));

    construct (L, &copy, 0);

    return 1;
}

int luaopen_core (lua_State *L)
{
    lua_newtable (L);

    /* Create an array object out of a Lua table or string. */
    
    lua_pushinteger (L, DOUBLE);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "doubles");
    
    lua_pushinteger (L, FLOAT);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "floats");
    
    lua_pushinteger (L, ULONG);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "ulongs");
    
    lua_pushinteger (L, LONG);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "longs");
    
    lua_pushinteger (L, UINT);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "uints");
    
    lua_pushinteger (L, INT);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "ints");
    
    lua_pushinteger (L, USHORT);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "ushorts");
    
    lua_pushinteger (L, SHORT);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "shorts");
    
    lua_pushinteger (L, UCHAR);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "uchars");
    
    lua_pushinteger (L, CHAR);
    lua_pushcclosure (L, create, 1);
    lua_setfield (L, -2, "chars");    

    /* Create arrays from other arrays. */
    
    lua_pushcfunction (L, copy);
    lua_setfield (L, -2, "copy");

    lua_pushcfunction (L, cast);
    lua_setfield (L, -2, "cast");
    
    return 1;
}
